#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "md3.h"

float anorms[256][256][3];

BOOL MD3_Load(char *filename, MD3_Model_t *MD3)
{
	FILE *stream;
	int i;
	long mesh_offset;

	if((stream=fopen(filename, "rb"))==NULL)
		return FALSE;

	fread(&MD3->header, sizeof(MD3_Header_t), 1, stream);

	if(MD3->header.magic!=MD3_MAGIC&&MD3->header.version!=15)
	{
		fclose(stream);
		return FALSE;
	}

	MD3->boneframes=malloc(sizeof(MD3_BoneFrame_t)*MD3->header.numBoneFrames);
	fread(MD3->boneframes, sizeof(MD3_BoneFrame_t), MD3->header.numBoneFrames, stream);

	MD3->tags=malloc(sizeof(MD3_Tag_t)*MD3->header.numBoneFrames*MD3->header.numTags);
	fread(MD3->tags, sizeof(MD3_Tag_t), MD3->header.numBoneFrames*MD3->header.numTags, stream);

	MD3->links=malloc(sizeof(MD3_Model_t)*MD3->header.numTags);

	for(i=0;i<MD3->header.numTags;i++)
		MD3->links[i]=NULL;

	MD3->meshes=malloc(sizeof(MD3_Mesh_t)*MD3->header.numMeshes);
	mesh_offset=ftell(stream);

	for(i=0;i<MD3->header.numMeshes;i++)
	{
		fseek(stream, mesh_offset, SEEK_SET);

		fread(&MD3->meshes[i].header, sizeof(MD3_Mesh_Header_t), 1, stream);

		MD3->meshes[i].skins=malloc(sizeof(MD3_Skin_t)*MD3->meshes[i].header.numSkins);
		fread(MD3->meshes[i].skins, sizeof(MD3_Skin_t), MD3->meshes[i].header.numSkins, stream);

		fseek(stream, mesh_offset+MD3->meshes[i].header.tri_start, SEEK_SET);
		MD3->meshes[i].triangles=malloc(sizeof(MD3_Triangle_t)*MD3->meshes[i].header.numTriangles);
		fread(MD3->meshes[i].triangles, sizeof(MD3_Triangle_t), MD3->meshes[i].header.numTriangles, stream);

		fseek(stream, mesh_offset+MD3->meshes[i].header.texvec_start, SEEK_SET);
		MD3->meshes[i].texcoords=malloc(sizeof(MD3_TexCoord_t)*MD3->meshes[i].header.numVertices);
		fread(MD3->meshes[i].texcoords, sizeof(MD3_TexCoord_t), MD3->meshes[i].header.numVertices, stream);

		fseek(stream, mesh_offset+MD3->meshes[i].header.vertex_start, SEEK_SET);
		MD3->meshes[i].vertices=malloc(sizeof(MD3_Vertex_t)*MD3->meshes[i].header.numVertices*MD3->meshes[i].header.numMeshFrames);
		fread(MD3->meshes[i].vertices, sizeof(MD3_Vertex_t), MD3->meshes[i].header.numMeshFrames*MD3->meshes[i].header.numVertices, stream);

		mesh_offset+=MD3->meshes[i].header.meshsize;
	}

	fclose(stream);

	{
		float a, b, c;
		int i, j;

		for(i=0;i<256;i++)
			for(j=0;j<256;j++)
			{
				float alpha=i*2.0f*3.1415f/255.0f;
				float beta=j*2.0f*3.1415f/255.0f;
				a=(float)(cos(beta)*sin(alpha));
				b=(float)(sin(beta)*sin(alpha));
				c=(float)cos(alpha);

				anorms[i][j][0]=a;
				anorms[i][j][1]=b;
				anorms[i][j][2]=c;
			}
	}

	return TRUE;
}

void MD3_DrawFrame(MD3_Model_t *MD3, int Frame, int NextFrame, float pol)
{
	int i, j, k, au, av;
	float v0[3], v1[3], n0[3], n1[3];

	for(k=0;k<MD3->header.numMeshes;k++)
	{
		if(MD3->meshes[k].texture)
			glBindTexture(GL_TEXTURE_2D, MD3->meshes[k].texid);

		for(i=0;i<MD3->meshes[k].header.numTriangles;i++)
		{
			glBegin(GL_TRIANGLES);

			for(j=0;j<3;j++)
			{
				v0[0]=(float)MD3->meshes[k].vertices[Frame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].vertex[0]/64.0f;
				v0[1]=(float)MD3->meshes[k].vertices[Frame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].vertex[1]/64.0f;
				v0[2]=(float)MD3->meshes[k].vertices[Frame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].vertex[2]/64.0f;

				v1[0]=(float)MD3->meshes[k].vertices[NextFrame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].vertex[0]/64.f;
				v1[1]=(float)MD3->meshes[k].vertices[NextFrame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].vertex[1]/64.f;
				v1[2]=(float)MD3->meshes[k].vertices[NextFrame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].vertex[2]/64.f;

				au=MD3->meshes[k].vertices[Frame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].normal[0];
				av=MD3->meshes[k].vertices[Frame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].normal[1];
				n0[0]=anorms[au][av][0];
				n0[1]=anorms[au][av][1];
				n0[2]=anorms[au][av][2];

				au=MD3->meshes[k].vertices[NextFrame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].normal[0];
				av=MD3->meshes[k].vertices[NextFrame*MD3->meshes[k].header.numVertices+MD3->meshes[k].triangles[i].vertex[j]].normal[1];
				n1[0]=anorms[au][av][0];
				n1[1]=anorms[au][av][1];
				n1[2]=anorms[au][av][2];

				glTexCoord2f(MD3->meshes[k].texcoords[MD3->meshes[k].triangles[i].vertex[j]].texcoord[0], -MD3->meshes[k].texcoords[MD3->meshes[k].triangles[i].vertex[j]].texcoord[1]);
				glNormal3f(n0[0]+pol*(n1[0]-n0[0]), n0[1]+pol*(n1[1]-n0[1]), n0[2]+pol*(n1[2]-n0[2]));
				glVertex3f(v0[0]+pol*(v1[0]-v0[0]), v0[1]+pol*(v1[1]-v0[1]), v0[2]+pol*(v1[2]-v0[2]));
			}

			glEnd();
		}
	}
}

void quat_from_matrix(float *quat, float *m)
{
	float tr, s, q[4];
	int i, j, k, nxt[3]={ 1, 2, 0 };

	tr=m[((0)*3)+(0)]+m[((1)*3)+(1)]+m[((2)*3)+(2)];

	if(tr>0.0f)
	{
		s=(float)sqrt(tr+1.0f);
		quat[3]=s/2.0f;
		s=0.5f/s;
		quat[0]=(m[((2)*3)+(1)]-m[((1)*3)+(2)])*s;
		quat[1]=(m[((0)*3)+(2)]-m[((2)*3)+(0)])*s;
		quat[2]=(m[((1)*3)+(0)]-m[((0)*3)+(1)])*s;
	}
	else
	{
		i=0;

		if(m[((1)*3)+(1)]>m[((0)*3)+(0)])
			i=1;

		if(m[((2)*3)+(2)]>m[((i)*3)+(i)])
			i=2;

		j=nxt[i];
		k=nxt[j];
		s=(float)sqrt((m[((i)*3)+(i)]-(m[((j)*3)+(j)]+m[((k)*3)+(k)]))+1.0f);
		q[i]=s*0.5f;

		if(s!=0.0f)
			s=0.5f/s;

		q[3]=(m[((k)*3)+(j)]-m[((j)*3)+(k)])*s;
		q[j]=(m[((j)*3)+(i)]+m[((i)*3)+(j)])*s;
		q[k]=(m[((k)*3)+(i)]+m[((i)*3)+(k)])*s;

		quat[0]=q[0];
		quat[1]=q[1];
		quat[2]=q[2];
		quat[3]=q[3];
	}
}

void quat_slerp(float *from, float *to, float t, float *res)
{
	float to1[4], omega, cosom, sinom, scale0, scale1;

	cosom=from[0]*to[0]+from[1]*to[1]+from[2]*to[2]+from[3]*to[3];

	if(cosom<0.0f)
	{
		cosom=-cosom;
		to1[0]=-to[0];
		to1[1]=-to[1];
		to1[2]=-to[2];
		to1[3]=-to[3];
	}
	else
	{
		to1[0]=to[0];
		to1[1]=to[1];
		to1[2]=to[2];
		to1[3]=to[3];
	}

	if((1.0f-cosom)>0.1f)
	{
		omega=(float)acos(cosom);
		sinom=(float)sin(omega);
		scale0=(float)sin((1.0f-t)*omega)/sinom;
		scale1=(float)sin(t*omega)/sinom;
	}
	else
	{
		scale0=1.0f-t;
		scale1=t;
	}

	res[0]=scale0*from[0]+scale1*to1[0];
	res[1]=scale0*from[1]+scale1*to1[1];
	res[2]=scale0*from[2]+scale1*to1[2];
	res[3]=scale0*from[3]+scale1*to1[3];
}

void matrix_from_quat(float *m, float *quat)
{
	float wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2;

	x2=quat[0]+quat[0]; 
	y2=quat[1]+quat[1]; 
	z2=quat[2]+quat[2];
	xx=quat[3]*x2;
	xy=quat[0]*y2;
	xz=quat[0]*z2;
	yy=quat[1]*y2;
	yz=quat[1]*z2;
	zz=quat[2]*z2;
	wx=quat[3]*x2;
	wy=quat[3]*y2;
	wz=quat[3]*z2;

	m[((0)*3)+(0)]=1.0f-(yy+zz);
	m[((1)*3)+(0)]=xy-wz;
	m[((2)*3)+(0)]=xz+wy;
	m[((0)*3)+(1)]=xy+wz;
	m[((1)*3)+(1)]=1.0f-(xx+zz);
	m[((2)*3)+(1)]=yz-wx;
	m[((0)*3)+(2)]=xz-wy;
	m[((1)*3)+(2)]=yz+wx;
	m[((2)*3)+(2)]=1.0f-(xx+yy);
}

void MD3_Draw(MD3_Model_t *MD3)
{
	int i;
	float m[16], quat0[4], quat1[4], quat[4], matrix[9], *position0, *matrix0, *position1, *matrix1;
	MD3_Model_t *link;

	MD3_DrawFrame(MD3, MD3->frame, MD3->nextframe, MD3->pol);

	for(i=0;i<MD3->header.numTags;i++)
	{
		link=MD3->links[i];

		if(link)
		{
			position0=MD3->tags[MD3->frame*MD3->header.numTags+i].position;
			matrix0=&MD3->tags[MD3->frame*MD3->header.numTags+i].rotation[0][0];
			position1=MD3->tags[MD3->nextframe*MD3->header.numTags+i].position;
			matrix1=&MD3->tags[MD3->nextframe*MD3->header.numTags+i].rotation[0][0];

			quat_from_matrix(quat0, matrix0);
			quat_from_matrix(quat1, matrix1);
			quat_slerp(quat0, quat1, MD3->pol, quat);
			matrix_from_quat(matrix, quat);

			m[0]=matrix[((0)*3)+(0)];
			m[1]=matrix[((1)*3)+(0)];
			m[2]=matrix[((2)*3)+(0)];
			m[3]=0.0f;
			m[4]=matrix[((0)*3)+(1)];
			m[5]=matrix[((1)*3)+(1)];
			m[6]=matrix[((2)*3)+(1)];
			m[7]=0.0f;
			m[8]=matrix[((0)*3)+(2)];
			m[9]=matrix[((1)*3)+(2)];
			m[10]=matrix[((2)*3)+(2)];
			m[11]=0.0f;
			m[12]=position0[0]+MD3->pol*(position1[0]-position0[0]);
			m[13]=position0[1]+MD3->pol*(position1[1]-position0[1]);
			m[14]=position0[2]+MD3->pol*(position1[2]-position0[2]);
			m[15]=1.0f;

			glPushMatrix();
			glMultMatrixf(m);
			MD3_Draw(link); // Mmm...  Recursiveness  :P~~~~
			glPopMatrix();
		}
	}
}

int MD3_LinkModel(char *tagname, MD3_Model_t *MD3, MD3_Model_t *Link)
{
	int i;

	for(i=0;i<MD3->header.numTags;i++)
		if(!strcmp(MD3->tags[i].name, tagname))
		{
			MD3->links[i]=Link;
			return i;
		}

	return -1;
}

int MD3_SetMeshTexture(char *meshname, MD3_Model_t *MD3, unsigned int texid)
{
	int i;

	for(i=0;i<MD3->header.numMeshes;i++)
	{
		if(!strcmp(MD3->meshes[i].header.name, meshname))
		{
			MD3->meshes[i].texid=texid;
			MD3->meshes[i].texture=1;
			return i;
		}
	}

	return -1;
}
