#define MD3_MAGIC ('I'+('D'<<8)+('P'<<16)+('3'<<24))
#define MD3_VERSION 15

typedef struct
{
	int magic;
	int version;
	char filename[68];
	int numBoneFrames;
	int numTags;
	int numMeshes;
	int numMaxSkins;
	int headerlength;
	int tag_start;
	int tag_end;
	int filesize;
} MD3_Header_t;

typedef struct
{
	float mins[3];
	float maxs[3];
	float position[3];
	float scale;
	char creator[16];
} MD3_BoneFrame_t;

typedef struct
{
	char name[64];
	float position[3];
	float rotation[3][3];
} MD3_Tag_t;

typedef struct
{
	int magic;
	char name[68];
	int numMeshFrames;
	int numSkins;
	int numVertices;
	int numTriangles;
	int tri_start;
	int headersize;
	int texvec_start;
	int vertex_start;
	int meshsize;
} MD3_Mesh_Header_t;

typedef struct
{
	char name[68];
} MD3_Skin_t;

typedef struct
{
	int vertex[3];
} MD3_Triangle_t;

typedef struct
{
	float texcoord[2];
} MD3_TexCoord_t;

typedef struct
{
	signed short vertex[3];
	unsigned char normal[2];
} MD3_Vertex_t;

typedef struct
{
	MD3_Mesh_Header_t header;
	MD3_Skin_t *skins;
	MD3_Triangle_t *triangles;
	MD3_TexCoord_t *texcoords;
	MD3_Vertex_t *vertices;
	int texture;
	unsigned int texid;
} MD3_Mesh_t;

typedef struct MD3_Model_s
{
	MD3_Header_t header;
	MD3_BoneFrame_t *boneframes;
	MD3_Tag_t *tags;
	MD3_Mesh_t *meshes;
	int frame, nextframe;
	float pol;
	struct MD3_Model_s **links;
} MD3_Model_t;

BOOL MD3_Load(char *filename, MD3_Model_t *MD3);
void MD3_DrawFrame(MD3_Model_t *MD3, int Frame, int NextFrame, float pol);
void MD3_Draw(MD3_Model_t *MD3);
int MD3_LinkModel(char *tagname, MD3_Model_t *MD3, MD3_Model_t *Link);
int MD3_SetMeshTexture(char *meshname, MD3_Model_t *MD3, unsigned int texid);
