#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include "tga.h"
#include "overlay.h"
#include "font.h"
//include 'md3.h' so that the code works
#include "md3.h"

HWND hWnd=NULL;
HDC hDC=NULL;
HGLRC hRC=NULL;

char *szAppName="OpenGL";

RECT WindowRect;

int Width=640, Height=480;
BOOL Done=FALSE, Key[255];

LARGE_INTEGER timerFrequency;
LARGE_INTEGER startTime;
LARGE_INTEGER currentTime;
float fps;
int nFrames=0;
int nLastFrames=0;
char OutBuffer[255];

// Display list, and texture ID holders
unsigned int FontBase, FontTexture, Base, Base1, Base2, Base3;

// MD3 models
MD3_Model_t lower, upper, head, weapon;

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void Render(void);
BOOL Init(void);
BOOL Create(void);
void Destroy(void);
void Resize(int Width, int Height);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int iCmdShow)
{
	WNDCLASS wc;
	MSG msg;

	wc.style=CS_VREDRAW|CS_HREDRAW|CS_OWNDC;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hInstance;
	wc.hIcon=LoadIcon(NULL, IDI_WINLOGO);
	wc.hCursor=LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground=GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName=NULL;
	wc.lpszClassName=szAppName;

	RegisterClass(&wc);

	WindowRect.left=0;
	WindowRect.right=Width;
	WindowRect.top=0;
	WindowRect.bottom=Height;

	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, FALSE);

	hWnd=CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW|WS_CLIPSIBLINGS, CW_USEDEFAULT, CW_USEDEFAULT, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInstance, NULL);

	ShowWindow(hWnd, SW_SHOW);
	SetForegroundWindow(hWnd);

	if(!Create())
		Done=TRUE;

	if(Done==FALSE)
		Resize(Width, Height);

	if(Done==FALSE)
		if(!Init())
			Done=TRUE;

	while(!Done)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if(msg.message==WM_QUIT)
				Done=TRUE;
			else
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
		{
			Render();
			SwapBuffers(hDC);

			nFrames++;
			if(nFrames>20)
			{
				QueryPerformanceCounter(&currentTime);
				fps=(float)nFrames/((float)(currentTime.QuadPart - startTime.QuadPart)/(float)timerFrequency.QuadPart);
				sprintf(OutBuffer, "FPS: %0.1f", fps);
				nFrames=0;
				QueryPerformanceCounter(&startTime);
			}
		}
	}

	Destroy();
	DestroyWindow(hWnd);
	return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_CREATE:
			QueryPerformanceFrequency(&timerFrequency);
			QueryPerformanceCounter(&startTime);
			break;

		case WM_CLOSE:
			PostQuitMessage(0);
			break;

		case WM_DESTROY:
			break;

		case WM_SIZE:
			Width=LOWORD(lParam);
			Height=HIWORD(lParam);
			Resize(Width, Height);
			break;

		case WM_KEYDOWN:
			Key[wParam]=TRUE;

			switch(wParam)
			{
				case VK_ESCAPE:
					PostQuitMessage(0);
					break;

				default:
					break;
			}
			break;

		case WM_KEYUP:
			Key[wParam]=FALSE;
			break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

void Render(void)
{

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	// Cull front facing triangles, only on the MD3 models
	glCullFace(GL_FRONT);

	// Enable texturing
	glEnable(GL_TEXTURE_2D);

	// Save the matrix
	glPushMatrix();
		// Reorient the model so that it faces twards the screen
		glTranslatef(0.0f, -10.0f, -50.0f);
		glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
		glRotatef(-90.0f, 0.0f, 0.0f, 1.0f);

		// Set what frames we are seing

		// Lower (legs) model (frame 100)
		lower.frame=100;
		lower.nextframe=100;
		lower.pol=0.0f;

		// Upper (torso) model (frame 100)
		upper.frame=100;
		upper.nextframe=100;
		upper.pol=0.0f;

		// Draw the lower model, which will recurse to the other linked models
		MD3_Draw(&lower);

	// Restore the matrix
	glPopMatrix();

	// Disable texturing
	glDisable(GL_TEXTURE_2D);

	// Reset culling, to cull back facing triangles
	glCullFace(GL_BACK);

	// Draw error status, and frame counter
	BeginOverlay(Width, Height);
		glColor3f(1.0f, 1.0f, 1.0f);
		Font_Print(0, 16, FontBase, FontTexture, "%s", gluErrorString(glGetError()));
		Font_Print(0, 0, FontBase, FontTexture, OutBuffer);
	EndOverlay();
}

BOOL Init(void)
{
	// Build font, and load font texture
	FontBase=Font_Build();
	FontTexture=TGA_UploadImage("font.tga");

	// Load the textures used by the models
	Base=TGA_UploadImage("base.tga");
	Base1=TGA_UploadImage("base1.tga");
	Base2=TGA_UploadImage("base2.tga");
	Base3=TGA_UploadImage("base3.tga");

	// Enable some states
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	// Load the models from disk
	MD3_Load("lower.md3", &lower);
	MD3_Load("upper.md3", &upper);
	MD3_Load("head.md3", &head);
	MD3_Load("rocketl.md3", &weapon);

	// Set the texture 'base.tga' to the mesh 'l_legs' on 'lower.md3'
	MD3_SetMeshTexture("l_legs", &lower, Base);

	// Set the texture 'base.tga' to the mesh 'u_arms' on 'upper.md3'
	MD3_SetMeshTexture("u_arms", &upper, Base);
	// Set the texture 'base1.tga' to the mesh 'u_chest' on 'upper.md3'
	MD3_SetMeshTexture("u_chest", &upper, Base1);

	// Set the texture 'base.tga' to the mesh 'h_helmet' on 'head.md3'
	MD3_SetMeshTexture("h_helmet", &head, Base);
	// Set no texture to the mesh 'h_visor' on 'lower.md3'
	MD3_SetMeshTexture("h_visor", &head, 0);

	// Set the texture 'base2.tga' to the mesh 'w_rocketl' on 'rocketl.md3'
	MD3_SetMeshTexture("w_rocketl", &weapon, Base2);
	// Set the texture 'base3.tga' to the mesh 'w_rocketl01' on 'rocketl.md3'
	MD3_SetMeshTexture("w_rocketl01", &weapon, Base3);

	// Link the models, so that they are recursively draw when MD3_Draw is called

	// Link 'upper.md3' to the tag 'tag_torso' on 'lower.md3'
	MD3_LinkModel("tag_torso", &lower, &upper);
	// Link 'head.md3' to the tag 'tag_head' on 'upper.md3'
	MD3_LinkModel("tag_head", &upper, &head);
	// Link 'rocketl.md3' to the tag 'tag_weapon' on 'upper.md3'
	MD3_LinkModel("tag_weapon", &upper, &weapon);

	return TRUE;
}

BOOL Create(void)
{
	PIXELFORMATDESCRIPTOR pfd;
	unsigned int PixelFormat;

	pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion=1;
	pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.cColorBits=32;
	pfd.cDepthBits=24;

	hDC=GetDC(hWnd);

	if(!(PixelFormat=ChoosePixelFormat(hDC, &pfd)))
	{
		MessageBox(hWnd, "ChoosePixelFormat Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!SetPixelFormat(hDC, PixelFormat, &pfd))
	{
		MessageBox(hWnd, "SetPixelFormat Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!(hRC=wglCreateContext(hDC)))
	{
		MessageBox(hWnd, "wglCreateContext Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	if(!wglMakeCurrent(hDC, hRC))
	{
		MessageBox(hWnd, "wglMakeCurrent Failed.", "ERROR", MB_OK);
		return FALSE;
	}

	return TRUE;
}

void Destroy(void)
{
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hRC);
	ReleaseDC(hWnd, hDC);
}

void Resize(int Width, int Height)
{
	if(Height==0)
		Height=1;

	glViewport(0,0, Width, Height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90.0f, (float)Width/Height, 0.001f, 400.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}
